package dev.fyloz.trial.colorrecipesexplorer.core.services;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fyloz.trial.colorrecipesexplorer.ColorRecipesExplorerApplication;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityAlreadyExistsException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.NullIdentifierException;
import dev.fyloz.trial.colorrecipesexplorer.core.model.IModel;
import org.slf4j.Logger;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.lang.NonNull;
import org.springframework.transaction.annotation.Transactional;

import javax.validation.constraints.NotNull;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

public class GenericService<T extends IModel, R extends JpaRepository<T, Long>> implements IGenericService<T> {

    protected Logger logger = ColorRecipesExplorerApplication.LOGGER;
    protected R dao;
    protected Class<T> type;

    public GenericService(R dao, Class<T> type) {
        this.dao = dao;
        this.type = type;
    }

    @Override
    public T getById(Long id) {
        Optional<T> found = dao.findById(id);
        if (found.isEmpty()) throw new EntityNotFoundException(type, EntityNotFoundException.IdentifierType.ID, id);

        return found.get();
    }

    @Override
    public List<T> getAll() {
        return dao.findAll();
    }

    @Override
    public T save(@NotNull T entity) {
        if (entity.getId() != null && existsById(entity.getId()))
            throw new EntityAlreadyExistsException(type, ModelException.IdentifierType.ID, entity.getId());

        return dao.save(entity);
    }

    @Override
    @Transactional
    public Collection<T> saveAll(@NotNull Collection<T> entities) {
        return entities
                .stream()
                .map(this::save)
                .collect(Collectors.toList());
    }

    @Override
    public T update(@NonNull T entity) {
        if (entity.getId() == null) throw new NullIdentifierException(type);
        if (!existsById(entity.getId()))
            throw new EntityNotFoundException(type, ModelException.IdentifierType.ID, entity.getId());

        return dao.save(entity);
    }

    @Override
    public void delete(@NonNull T entity) {
        if (entity.getId() == null) throw new NullIdentifierException(type);
        if (!existsById(entity.getId()))
            throw new EntityNotFoundException(type, ModelException.IdentifierType.ID, entity.getId());

        dao.delete(entity);
    }

    @Override
    public void deleteById(Long id) {
        if (!existsById(id))
            throw new EntityNotFoundException(type, ModelException.IdentifierType.ID, id);

        delete(getById(id));
    }

    @Override
    @Transactional
    public void deleteAll(@NonNull List<T> entities) {
        entities
                .forEach(this::delete);
    }

    /**
     * Vérifie si une entité est valide pour la création dans la base de données.
     *
     * @param entity L'entité à vérifier.
     * @return Si l'entité est valide pour la création.
     */
    @Deprecated(since = "1.3.0", forRemoval = true)
    public boolean isValidForCreation(T entity) {
        return entity != null && !existsById(entity.getId());
    }

    /**
     * Vérifie si une entité est valide pour la création dans la base de données.
     *
     * @param entity L'entité à vérifier.
     * @return Si l'entité est valide pour l'édition.
     */
    @Deprecated(since = "1.3.0", forRemoval = true)
    public boolean isValidForUpdate(@NonNull T entity) {
        return entity.getId() != null && existsById(entity.getId());
    }

    @Override
    @Deprecated(since = "v.1.3.0", forRemoval = true)
    public boolean exists(T entity) {
        return entity != null && entity.getId() != null && existsById(entity.getId());
    }

    @Override
    public final boolean existsById(Long id) {
        return dao.existsById(id);
    }

    /**
     * Transforme un objet en Json.
     *
     * @param obj L'objet à transformer
     * @return L'objet sous forme de Json.
     */
    public String asJson(Object obj) {
        ObjectMapper objectMapper = new ObjectMapper();
        try {
            return objectMapper.writeValueAsString(obj);
        } catch (JsonProcessingException e) {
            logger.error("Une erreur est survenue lors de la transformation d'un objet en Json", e);
            return null;
        }
    }
}
