package dev.fyloz.trial.colorrecipesexplorer.core.services.model;

import dev.fyloz.trial.colorrecipesexplorer.core.io.response.ModelResponseBuilder;
import dev.fyloz.trial.colorrecipesexplorer.core.io.response.ResponseCode;
import dev.fyloz.trial.colorrecipesexplorer.core.model.*;
import dev.fyloz.trial.colorrecipesexplorer.core.model.dto.MixCreationFormDto;
import dev.fyloz.trial.colorrecipesexplorer.core.services.GenericService;
import dev.fyloz.trial.colorrecipesexplorer.dao.MixDao;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.validation.constraints.NotNull;
import java.util.*;
import java.util.stream.Collectors;

import static dev.fyloz.trial.colorrecipesexplorer.web.StringBank.MATERIALS;

@Service
public class MixService extends GenericService<Mix, MixDao> {

    private MaterialService materialService;
    private MixQuantityService mixQuantityService;
    private MixTypeService mixTypeService;
    private RecipeService recipeService;

    @Autowired
    public MixService(MixDao mixDao, MaterialService materialService, MixQuantityService mixQuantityService, MixTypeService mixTypeService, RecipeService recipeService) {
        super(mixDao, Mix.class);
        this.materialService = materialService;
        this.mixQuantityService = mixQuantityService;
        this.mixTypeService = mixTypeService;
        this.recipeService = recipeService;
    }

    /**
     * Récupère les produits disponibles pour un mélange.
     * Le mélange peut ne pas exister.
     *
     * @param recipeId L'identifiant de la recette dans laquelle se trouve le mélange
     * @param mixId    L'identifiant du mélange (-1 si le mélange n'existe pas)
     * @return Les produits disponibles pour ce mélange
     */
    public Collection<Material> getAvailableMaterialsForMixId(Long recipeId, Long mixId) {
        return existsById(mixId) ? getAvailableMaterialsForMix(getById(mixId)) : getAvailableMaterialsForNewMix(recipeService.getById(recipeId));
    }

    /**
     * Récupère les produits disponibles pour un mélange existant.
     *
     * @param mix Le mélange
     * @return Les produits disponibles pour ce mélange
     */
    public Collection<Material> getAvailableMaterialsForMix(Mix mix) {
        return getAvailableMaterialsForNewMix(mix.getRecipe())
                .stream()
                .filter(m -> !m.equals(mix.getMixType().getMaterial()))
                .collect(Collectors.toList());
    }

    /**
     * Récupère les produits disponibles pour un mélange inexistant.
     *
     * @param recipe La recette dans laquelle se trouve le mélange
     * @return Les produits disponibles pour le nouveau mélange
     */
    public Collection<Material> getAvailableMaterialsForNewMix(Recipe recipe) {
        Collection<MixType> recipeMixTypes = recipeService.getAssociatedMixesTypes(recipe);

        return materialService
                .getAll()
                .stream()
                .filter(m -> !m.isMixType() || recipeMixTypes.contains(mixTypeService.getByMaterial(m).get()))
                .sorted(Comparator.comparing(Material::getName))
                .sorted(Comparator.comparing(m -> m.getMaterialType().getName()))
                .collect(Collectors.toList());
    }

    @Transactional
    public ModelResponseBuilder create(MixCreationFormDto formDto, @NotNull Recipe recipe) {
        ModelResponseBuilder modelResponseBuilder = new ModelResponseBuilder();

        List<Material> materials = new ArrayList<>();
        for (String materialCode : formDto.getMaterials()) {
            Optional<Material> found = materialService.getByName(materialCode);
            if (found.isEmpty()) {
                return modelResponseBuilder.addResponseCode(ResponseCode.MATERIAL_NOT_FOUND_BY_NAME, materialCode);
            }

            materials.add(found.get());
        }

        Optional<MixType> optionalMixType = mixTypeService.createByName(formDto.getMixTypeName(), formDto.getMaterialType());
        if (optionalMixType.isEmpty()) return modelResponseBuilder.addResponseCode(ResponseCode.ERROR_SAVING);

        MixType mixType = optionalMixType.get();
        if (recipeService.hasMixType(recipe, mixType))
            return modelResponseBuilder.addResponseCode(ResponseCode.MIX_TYPE_ALREADY_USED, mixType.getName());

        // Crée le mélange en premier pour avoir accès à son ID pour les autres éléments
        Mix mix = new Mix(recipe, mixType);

        Optional<Mix> savedMix = save(mix);
        if (savedMix.isPresent()) {
            mix = savedMix.get();

            List<MixQuantity> mixQuantities = createMixQuantities(savedMix.get(), materials, formDto.getQuantities());
            mix.setMixQuantities(mixQuantities);

            // Retourne aucune erreur s'il la mise à jour à lieu
            if (update(mix).isPresent()) return null;

            deleteMix(mix);
        }

        return modelResponseBuilder.addResponseCode(ResponseCode.ERROR_SAVING);
    }

    @Transactional
    public ModelResponseBuilder edit(Mix mix, MixCreationFormDto formDto) {
        ModelResponseBuilder modelResponseBuilder = new ModelResponseBuilder();
        Material material = mix.getMixType().getMaterial();

        List<Material> materials = new ArrayList<>();
        for (String materialCode : formDto.getMaterials()) {
            Optional<Material> found = materialService.getByName(materialCode);
            if (found.isEmpty()) {
                return modelResponseBuilder.addResponseCode(ResponseCode.MATERIAL_NOT_FOUND_BY_NAME, materialCode);
            }

            materials.add(found.get());
        }

        mix.getMixType().getMaterial().setMaterialType(formDto.getMaterialType());
        mix.getMixType().setName(formDto.getMixTypeName());
        material.setName(formDto.getMixTypeName());

        List<MixQuantity> oldQuantities = mix.getMixQuantities();
        List<MixQuantity> mixQuantities = createMixQuantities(mix, materials, formDto.getQuantities());

        // Supprime les anciens MixQuantity pour éviter les doublons et les entrées inutiles dans la base de données
        if (!mixQuantityService.deleteAll(oldQuantities)) {
            return modelResponseBuilder.addResponseCode(ResponseCode.ERROR_SAVING);
        }

        mix.setMixQuantities(mixQuantities);
        if (materialService.update(material).isPresent() && update(mix).isPresent()) {
            return null;
        } else {
            return modelResponseBuilder.addResponseCode(ResponseCode.ERROR_SAVING);
        }
    }

    @Deprecated(since = "1.3.0", forRemoval = true)
    public void deleteMix(Mix mix) {
        mixQuantityService.deleteAll(mix.getMixQuantities());
        delete(mix);
    }

    @Override
    public void delete(Mix mix) {
        mixQuantityService.deleteAll(mix.getMixQuantities());
/
        super.delete(mix);
    }

    @Override
    public void deleteById(Long id) {
        delete(getById(id));
    }

    /**
     * Crée une liste de MixQuantity, pour la création d'un mélange
     *
     * @param mix        Le mélange associé aux MixQuantity
     * @param materials  Les matériaux
     * @param quantities Les quantités
     * @return La liste des MixQuantity créés.
     */
    public List<MixQuantity> createMixQuantities(Mix mix, List<Material> materials, List<Float> quantities) {
        List<MixQuantity> mixQuantities = new ArrayList<>();

        for (int i = 0; i < materials.size(); i++) {
            Material material = materials.get(i);
            float quantity = quantities.get(i);

            MixQuantity mixQuantity = new MixQuantity(mix, material, quantity);
            mixQuantities.add(mixQuantity);
        }

        return mixQuantities;
    }
}
