package dev.fyloz.trial.colorrecipesexplorer.core.services.model;

import dev.fyloz.trial.colorrecipesexplorer.core.io.file.FileHandler;
import dev.fyloz.trial.colorrecipesexplorer.core.io.file.ImageHandler;
import dev.fyloz.trial.colorrecipesexplorer.core.model.*;
import dev.fyloz.trial.colorrecipesexplorer.core.model.dto.RecipeEditorFormDto;
import dev.fyloz.trial.colorrecipesexplorer.core.model.dto.RecipeExplorerFormDto;
import dev.fyloz.trial.colorrecipesexplorer.core.services.GenericService;
import dev.fyloz.trial.colorrecipesexplorer.dao.RecipeDao;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.MultiValueMap;

import java.io.File;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class RecipeService extends GenericService<Recipe, RecipeDao> {

    private CompanyService companyService;
    private MixService mixService;
    private StepService stepService;

    @Autowired
    public RecipeService(RecipeDao recipeDao, CompanyService companyService, MixService mixService, StepService stepService) {
        super(recipeDao, Recipe.class);
        this.companyService = companyService;
        this.mixService = mixService;
        this.stepService = stepService;
    }

    /**
     * Récupère la liste des recettes associées à une compagnie.
     *
     * @param company La compagnie.
     * @return La liste des recettes associées à cette compagnie.
     */
    public List<Recipe> getByCompany(Company company) {
        return dao.findAllByCompany(company);
    }

    /**
     * Récupère une liste triée des mélanges pour une recette.
     *
     * @param recipe La recette dont on veut récupérer les mélanges
     * @return Une liste triée des mélanges.
     */
    public List<Mix> getSortedMixes(Recipe recipe) {
        List<Mix> mixes = recipe.getMixes();
        mixes.sort(Comparator.comparing(Mix::getId));
        return new ArrayList<>(mixes);  // Convertit le PersistentBag en ArrayList
    }

    /**
     * Récupère toutes les recettes dans la base de données et les classes par compagnie.
     *
     * @return Un Map contenant les recettes classées par compagnie.
     */
    public Map<Company, List<Recipe>> getRecipesByCompany() {
        return mappedByCompany(getAll());
    }

    @Deprecated(since = "1.3.0", forRemoval = true)
    public Recipe updateRecipe(Recipe newRecipe, Recipe storedRecipe, MultiValueMap<String, Object> form) {
        storedRecipe.setName(newRecipe.getName());
        storedRecipe.setCompany(newRecipe.getCompany());
        storedRecipe.setDescription(newRecipe.getDescription());
        storedRecipe.setSample(newRecipe.getSample());
        storedRecipe.setApprobationDate(newRecipe.getApprobationDate());
        storedRecipe.setRemark(newRecipe.getRemark());
        storedRecipe.setNote(newRecipe.getNote());

        return convertAndCreateSteps(storedRecipe, form);
    }

    /**
     * Met à jour une recette ainsi que ses étapes.
     *
     * @param recipeDto Les informations de la recette à mettre à jour
     * @return La recette mise à jour
     */
    @Transactional
    public Recipe updateRecipeAndSteps(RecipeEditorFormDto recipeDto) {
        Recipe recipe = recipeDto.getRecipe();

        stepService.createAllForRecipe(recipe, recipeDto.getStep());
        return update(recipeDto.getRecipe());
    }

    /**
     * Met à jour les informations d'une recette trouvées dans l'explorateur de recette.
     *
     * @param form Le formulaire contenant les données mises à jour
     */
    @Transactional
    public void updateRecipeExplorerInfos(RecipeExplorerFormDto form) {
        long recipeId = form.getRecipeId();
        Map<Long, String> locations = form.getLocations();
        String note = form.getNote();

        Recipe recipe = getById(recipeId);

        // Note
        recipe.setNote(note);
        save(recipe);

        // Casiers
        for (Map.Entry<Long, String> location : locations.entrySet()) {
            Mix mix = mixService.getById(location.getKey());
            mix.setLocation(location.getValue());
            mixService.update(mix);
        }
    }

    @Override
    public void delete(Recipe recipe) {
        super.delete(recipe);
        removeAllFiles(recipe);
    }

    /**
     * Récupère le nom des images liées à une recette.
     *
     * @param recipe La recette dont on veut récupérer les images
     * @return Une liste contenant le nom des images liées à la recette.
     */
    public List<String> getImageFiles(Recipe recipe) {
        Long recipeId = recipe.getId();
        String name = recipe.getName();
        String fileName = String.format("%s_%s", recipeId, name);

        File imageLocation = new File(ImageHandler.IMAGES_LOCATION);
        File[] result = imageLocation.listFiles((d, n) -> n.startsWith(fileName) && n.endsWith("jpeg"));

        if (result == null) return new ArrayList<>();

        return Arrays.stream(result).map(File::getName).collect(Collectors.toList());
    }

    /**
     * Récupère les mélanges d'une recette triés selon leur identifiant.
     *
     * @param recipe La recette
     * @return Les mélanges de la recette triés selon leur identifiant
     */
    public Collection<Mix> getMixesSortedById(Recipe recipe) {
        return getMixesSorted(recipe, Comparator.comparing(Mix::getId));
    }

    /**
     * Récupère les mélanges d'une recette triés selon un comparateur.
     *
     * @param recipe     La recette
     * @param comparator Le comparateur, ou la méthode de triage
     * @return Les mélanges de la recette triés selon le comparateur
     */
    public Collection<Mix> getMixesSorted(Recipe recipe, Comparator<? super Mix> comparator) {
        return recipe
                .getMixes()
                .parallelStream()
                .sorted(comparator)
                .collect(Collectors.toList());
    }

    /**
     * Récupère les types de mélanges associés à la recette.
     *
     * @param recipe La recette dont on veut récupérer les types de mélange
     * @return Une liste contenant les types de mélanges.
     */
    public List<MixType> getAssociatedMixesTypes(Recipe recipe) {
        return recipe
                .getMixes()
                .stream()
                .map(Mix::getMixType)
                .collect(Collectors.toList());
    }

    /**
     * Vérifie si une recette contient un mélange ayant le type de mélange spécifié.
     *
     * @param recipe  La recette
     * @param mixType Le type de mélange
     * @return Si la recette contient le type de mélange
     */
    public boolean hasMixType(Recipe recipe, MixType mixType) {
        return getAssociatedMixesTypes(recipe)
                .contains(mixType);
    }

    private Map<Company, List<Recipe>> mappedByCompany(List<Recipe> recipes) {
        List<Company> companies = companyService.getAll();
        Map<Company, List<Recipe>> mappedRecipes = new HashMap<>();

        for (Company c : companies) {
            List<Recipe> recipesForCompany = recipes.stream().filter(r -> r.getCompany().equals(c)).collect(Collectors.toList());

            if (recipesForCompany.size() > 0) {
                mappedRecipes.put(c, recipesForCompany);
            }
        }

        return mappedRecipes;
    }

    /**
     * Crée et assigne des étapes à une recette, en se basant sur l'entrée de utilisateur.
     *
     * @param recipe La recette à assigner les étapes
     * @param input  L'entrée de l'utilisateur, contenant les étapes (key = "step_*")
     * @return La recette avec les étapes.
     */
    @Transactional
    public Recipe convertAndCreateSteps(Recipe recipe, MultiValueMap<String, Object> input) {
        // Convertit les étapes en RecipeSteps
        List<RecipeStep> steps = new ArrayList<>();
        if (input.containsKey("step")) {
            Object field = input.get("step");

            if (field instanceof LinkedList && ((LinkedList) field).getFirst() instanceof String) {
                for (String step : (LinkedList<String>) field) {
                    steps.add(new RecipeStep(recipe, step));
                }
            }
        }

        return setSteps(recipe, steps);
    }

    private Recipe setSteps(Recipe recipe, List<RecipeStep> steps) {
        stepService.deleteAll(recipe.getRecipeSteps());

        recipe.setRecipeSteps(steps);
        return update(recipe);
    }

    private boolean removeAllFiles(Recipe recipe) {
        FileHandler fileHandler;
        for (String image : getImageFiles(recipe)) {
            fileHandler = new FileHandler(image.replace(".jpeg", ""), FileHandler.FileContext.IMAGE, FileHandler.FileExtension.JPEG);

            if (!fileHandler.deleteFile()) return false;
        }

        return true;
    }

    public String getImageFileName(Recipe recipe) {
        return String.format("%s_%s", recipe.getId(), recipe.getName());
    }
}
