package dev.fyloz.trial.colorrecipesexplorer.core.services;

import dev.fyloz.trial.colorrecipesexplorer.core.model.Material;
import dev.fyloz.trial.colorrecipesexplorer.core.model.MaterialType;
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MaterialService;
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MixQuantityService;
import dev.fyloz.trial.colorrecipesexplorer.dao.MaterialDao;
import org.junit.Before;
import org.junit.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.mockito.Spy;

import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doReturn;

public class MaterialServiceTests {

    @Mock
    private MaterialDao materialDao;

    @Mock
    private MixQuantityService mixQuantityService;

    @Spy
    @InjectMocks
    private MaterialService materialService;

    private Material testMaterial;
    private MaterialType testMaterialType;
    private List<Material> testMaterialsList;

    @Before
    public void init() {
        MockitoAnnotations.initMocks(this);

        testMaterial = new Material();
        testMaterial.setMaterialID(0);
        testMaterial.setName("Test Material");

        testMaterialType = new MaterialType();
        testMaterialType.setMaterialTypeID(0);
        testMaterialType.setMaterialTypeName("Test Type");

        testMaterialsList = new ArrayList<>();
        testMaterialsList.add(testMaterial);
    }

    @Test
    public void getAllByMaterialTypeTest() {
        // Arrange
        doReturn(testMaterialsList).when(materialDao).findAllByMaterialType(testMaterialType);

        // Act
        List<Material> actual = materialService.getAllByMaterialType(testMaterialType);

        // Assert
        assertTrue(actual.contains(testMaterial));
    }

    @Test
    public void getAllByMaterialTypeTestShouldBeEmpty() {
        // Arrange
        doReturn(new ArrayList<>()).when(materialDao).findAllByMaterialType(testMaterialType);

        // Act
        List<Material> actual = materialService.getAllByMaterialType(testMaterialType);

        // Assert
        assertTrue(actual.isEmpty());
    }

    @Test
    public void getAllByMaterialTypeTestNull() {
        // Arrange
        doReturn(testMaterialsList).when(materialDao).findAllByMaterialType(any());

        // Act
        List<Material> actual = materialService.getAllByMaterialType(null);

        // Assert
        assertTrue(actual.isEmpty());
    }

    @Test
    public void isLinkedToMixesTestShouldBeTrue() {
        // Arrange
        doReturn(true).when(mixQuantityService).existsByMaterial(testMaterial);

        // Act
        boolean actual = materialService.isLinkedToMixes(testMaterial);

        // Arrange
        assertTrue(actual);
    }

    @Test
    public void isLinkedToMixesTestShouldBeFalse() {
        // Arrange
        doReturn(false).when(mixQuantityService).existsByMaterial(testMaterial);

        // Act
        boolean actual = materialService.isLinkedToMixes(testMaterial);

        // Arrange
        assertFalse(actual);
    }

    @Test
    public void isLinkedToMixesTestNull() {
        // Arrange
        doReturn(false).when(mixQuantityService).existsByMaterial(any());

        // Act
        boolean actual = materialService.isLinkedToMixes(null);

        // Arrange
        assertFalse(actual);
    }

    @Test
    public void deleteIfNotLinkedTestShouldBeTrue() {
        // Arrange
        doReturn(false).when(materialService).isLinkedToMixes(testMaterial);
        doReturn(true).when(materialService).delete(testMaterial);

        // Act
        boolean actual = materialService.deleteIfNotLinked(testMaterial);

        // Assert
        assertTrue(actual);
    }

    @Test
    public void deleteIfNotLinkedTestShouldBeFalse() {
        // Arrange
        doReturn(true).when(materialService).isLinkedToMixes(testMaterial);
        doReturn(false).when(materialService).delete(testMaterial);

        // Act
        boolean actual = materialService.deleteIfNotLinked(testMaterial);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void deleteIfNotLinkedTestShouldBeFalseNotLinked() {
        // Arrange
        doReturn(false).when(materialService).isLinkedToMixes(testMaterial);
        doReturn(false).when(materialService).delete(testMaterial);

        // Act
        boolean actual = materialService.deleteIfNotLinked(testMaterial);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void deleteIfNotLinkedTestNull() {
        // Arrange
        doReturn(false).when(materialService).isLinkedToMixes(testMaterial);
        doReturn(true).when(materialService).delete(testMaterial);

        // Act
        boolean actual = materialService.deleteIfNotLinked(null);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void existsTestShouldBeTrue() {
        // Arrange
        doReturn(true).when((GenericService<Material, MaterialDao>) materialService).exists(testMaterial);
        doReturn(true).when(materialDao).existsByMaterialCode(testMaterial.getName());

        // Act
        boolean actual = materialService.exists(testMaterial);

        // Assert
        assertTrue(actual);
    }

    @Test
    public void existsTestShouldBeFalse() {
        // Arrange
        doReturn(false).when((GenericService<Material, MaterialDao>) materialService).exists(testMaterial);
        doReturn(false).when(materialDao).existsByMaterialCode(testMaterial.getName());

        // Act
        boolean actual = materialService.exists(testMaterial);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void existsTestShouldBeTrueOneConditionMet() {
        // Arrange
        doReturn(true).when(materialDao).existsByMaterialCode(testMaterial.getName());

        // Act
        boolean actual = materialService.exists(testMaterial);

        // Assert
        assertTrue(actual);
    }

    @Test
    public void existsTestNull() {
        // Arrange
        doReturn(true).when(materialDao).existsByMaterialCode(any());

        // Act
        boolean actual = materialService.exists(null);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void isValidForUpdateTestShouldBeTrue() {
        // Arrange
        doReturn(null).when(materialDao).findByMaterialCode(testMaterial.getName());
        doReturn(true).when((GenericService<Material, MaterialDao>) materialService).isValidForUpdate(testMaterial);

        // Act
        boolean actual = materialService.isValidForUpdate(testMaterial);

        // Assert
        assertTrue(actual);
    }

    @Test
    public void isValidForUpdateTestShouldBeTrueMaterialCodeExistsWithSameId() {
        // Arrange
        doReturn(testMaterial).when(materialDao).findByMaterialCode(testMaterial.getName());
        doReturn(true).when((GenericService<Material, MaterialDao>) materialService).isValidForUpdate(testMaterial);

        // Act
        boolean actual = materialService.isValidForUpdate(testMaterial);

        // Assert
        assertTrue(actual);
    }

    @Test
    public void isValidForUpdateTestShouldBeFalse() {
        // Arrange
        doReturn(testMaterial).when(materialDao).findByMaterialCode(testMaterial.getName());
        doReturn(false).when((GenericService<Material, MaterialDao>) materialService).isValidForUpdate(testMaterial);

        // Act
        boolean actual = materialService.isValidForUpdate(testMaterial);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void isValidForUpdateTestShouldBeFalseMaterialCodeExists() {
        // Arrange
        Material otherMaterial = new Material();
        otherMaterial.setName(testMaterial.getName());
        otherMaterial.setMaterialID(1);

        doReturn(otherMaterial).when(materialDao).findByMaterialCode(testMaterial.getName());

        // Act
        boolean actual = materialService.isValidForUpdate(testMaterial);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void isValidForUpdateShouldNull() {
        // Arrange
        doReturn(testMaterial).when(materialDao).findByMaterialCode(testMaterial.getName());

        // Act
        boolean actual = materialService.isValidForUpdate(null);

        // Assert
        assertFalse(actual);
    }

    @Test
    public void getAllBySearchStringTest() {
        // Arrange
        doReturn(testMaterialsList).when(materialDao).findAllByMaterialCodeContainingIgnoreCase(testMaterial.getName());

        // Act
        List<Material> actual = materialService.getAllBySearchString(testMaterial.getName());

        // Assert
        assertTrue(actual.contains(testMaterial));
    }

    @Test
    public void getAllBySearchStringTestShouldBeEmpty() {
        // Arrange
        doReturn(new ArrayList<>()).when(materialDao).findAllByMaterialCodeContainingIgnoreCase(testMaterial.getName());

        // Act
        List<Material> actual = materialService.getAllBySearchString(testMaterial.getName());

        // Assert
        assertTrue(actual.isEmpty());
    }

    @Test
    public void getAllBySearchStringTestMixType() {
        // Arrange
        Material mixTypeMaterial = new Material();
        mixTypeMaterial.setMaterialID(2);
        mixTypeMaterial.setName("Mix Type");
        mixTypeMaterial.setMixType(true);

        testMaterialsList.add(mixTypeMaterial);

        doReturn(testMaterialsList).when(materialDao).findAllByMaterialCodeContainingIgnoreCase(mixTypeMaterial.getName());

        // Act
        List<Material> actual = materialService.getAllBySearchString(mixTypeMaterial.getName());

        // Assert
        assertFalse(actual.contains(mixTypeMaterial));
    }

}
